#include <windows.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include "modcliprc.h"

#define WM_TASKTRAY		(WM_USER+0x1000)
#define TASKTRAY_ID		(0)
#define IDM_QUIT		(0x100)
#define MODCLIP_WINDOW_NAME	"modclip main window"
#define MODCLIP_WINDOW_CLASS	"modclip main window"

HINSTANCE hInst;
HMENU hMenu;
HWND hwndNextViewer = NULL;

static int
set_tasktray_icon(HWND hwnd)
{
    NOTIFYICONDATA nid;
    BOOL ret;

    memset(&nid, 0, sizeof (nid));
    nid.cbSize = sizeof (nid);
    nid.hWnd = hwnd;
    nid.uID = TASKTRAY_ID;
    nid.uFlags = NIF_ICON|NIF_MESSAGE|NIF_TIP;
    nid.uCallbackMessage = WM_TASKTRAY;
    nid.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_ICON));
    strcpy(nid.szTip, "clipboard modifier");
    ret = Shell_NotifyIcon(NIM_ADD, &nid);
    if (!ret)
	MessageBox(NULL, "cannot set task tray icon.", "Error", MB_OK);
    return !ret;
}

static void
remove_tasktray_icon(HWND hwnd)
{
    NOTIFYICONDATA nid;

    memset(&nid, 0, sizeof (nid));
    nid.cbSize = sizeof (nid);
    nid.hWnd = hwnd;
    nid.uID = TASKTRAY_ID;
    Shell_NotifyIcon(NIM_DELETE, &nid);
}

static int
create_tasktray_menu(HWND hwnd)
{
    hMenu = CreatePopupMenu();
    if (!hMenu)
	goto error;
    if (!AppendMenu(hMenu, MF_STRING, IDM_QUIT, "&Quit"))
	goto error;
    return 0;
error:
    MessageBox(NULL, "cannot create task tray menu.", "Error", MB_OK);
    return -1;
}

LRESULT
TaskTrayProc(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = 0;
    POINT point;
    RECT rect = { 0, 0, 0, 0 };
    HWND hwndShell;
    TPMPARAMS tpmp;

    switch (lParam) {
    case WM_RBUTTONDOWN:
	memset(&tpmp, 0, sizeof (tpmp));
	hwndShell = FindWindow("Shell_TrayWnd", NULL);
	if (hwndShell)
	    GetWindowRect(hwndShell, &rect);
	tpmp.cbSize = sizeof (tpmp);
	tpmp.rcExclude = rect;

	GetCursorPos(&point);
	SetForegroundWindow(hwnd);

	TrackPopupMenuEx(hMenu,
			 TPM_LEFTALIGN|TPM_LEFTBUTTON, point.x, point.y,
			 hwnd, &tpmp);
	break;
    }

    return ret;
}

BOOL
skip_spaces(LPSTR *psrc)
{
    BOOL skipped = FALSE;

    assert(psrc != NULL && *psrc != NULL);
    while (**psrc && isspace(**psrc)) {
	skipped = TRUE;
	(*psrc)++;
    }

    return skipped;
}

BOOL
ensure_to_terminate(LPSTR *pdst, size_t *pdstmax)
{
    assert(*pdstmax > 0);
    **pdst = '\0';
    return *pdstmax == 1;
}

BOOL
put_char(LPSTR *pdst, size_t *pdstmax, char ch)
{
    BOOL end;

    assert(*pdstmax > 0);
    end = *pdstmax <= 1;
    if (!end) {
	**pdst = ch;
	(*pdst)++;
	(*pdstmax)--;
    }
    **pdst = '\0'; // ensure to terminate.
    return end;
}

BOOL
clean_up_string(LPSTR dst, size_t dstmax, LPSTR src)
{
    BOOL modified = FALSE, newline = FALSE;
    int num_backslash;

    num_backslash = 0;
    if (ensure_to_terminate(&dst, &dstmax))
	return modified;
    while (*src) {
	if (ensure_to_terminate(&dst, &dstmax))
	    return modified;

	newline = (*src == '\r' || *src == '\n');
	if (isspace(*src) && !newline) {
	    // trailing spaces is eliminated.
	    modified = TRUE;
	    goto quit;
	}
	if (*src == '\\') {
	    // mule may append multiple backslashes for line continuation.
	    num_backslash++;
	    goto next;
	}
	if (newline) {
	    modified = TRUE;
	    num_backslash = 0;
	    goto next;
	}
	// dump suspended backslashes.
	while (num_backslash > 0) {
	    if (put_char(&dst, &dstmax, '\\'))
		goto quit;
	    num_backslash--;
	}
	if (put_char(&dst, &dstmax, *src))
	    goto quit;
next:
	src++;
    }
quit:
    return modified;
}

static HGLOBAL
modify_string(LPSTR src)
{
    // XXX: hardwired. :-<
    BOOL match = FALSE;
    BOOL skipped = FALSE;
    BOOL need_h = FALSE;
    BOOL modified = FALSE;
    char buf[10];
#define PREFIX1 "http://"
#define PREFIX2 "ttp://"
#define GETLEN(p) (sizeof ("" p "") -1)

    skipped = skip_spaces(&src);

    clean_up_string(buf, sizeof (buf), src);
    if (!memcmp(buf, PREFIX1, GETLEN(PREFIX1))) {
	match = TRUE;
    } else if (!memcmp(buf, PREFIX2, GETLEN(PREFIX2))) {
	match = TRUE;
	need_h = TRUE;
    }
    if (match) {
	HGLOBAL hGlobal;
	LPSTR outstr, p;
	size_t len;

	len = strlen(src) + 1 /* for 'h' */ + 1 /* for \0 */;
	hGlobal = GlobalAlloc(GMEM_MOVEABLE, len);
	outstr = p = GlobalLock(hGlobal);
	if (need_h) {
	    *p++ = 'h';
	    len--;
	}
	modified = clean_up_string(p, len, src);
	GlobalUnlock(hGlobal);
	if (modified || skipped)
	    return hGlobal;
	GlobalFree(hGlobal);
    }

    return NULL;
}

#define NumOf(array)	(sizeof (array) / sizeof (array[0]))
LRESULT
DrawClipboardProc(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = 0;
    int id;
    UINT formats[] = { CF_TEXT };
    HWND hwndowner;

    hwndowner = GetClipboardOwner();
    if (hwndowner == hwnd)
	return 0;

    id = GetPriorityClipboardFormat(formats, NumOf(formats));
    switch (id) {
    case CF_TEXT:
	if (OpenClipboard(hwnd)) {
	    HGLOBAL hGlobal, hGlbModified;
	    LPSTR str;
	    hGlobal = GetClipboardData(id);
	    str = GlobalLock(hGlobal);
	    hGlbModified = modify_string(str);
	    GlobalUnlock(hGlobal);
	    if (hGlbModified) {
		EmptyClipboard();
		SetClipboardData(CF_TEXT, hGlbModified);
	    }
	    CloseClipboard();
	}
	break;
    }

    return ret;
}

int
initialize(HWND hwnd)
{
    if (set_tasktray_icon(hwnd))
	return -1;
    if (create_tasktray_menu(hwnd))
	return -1;
    hwndNextViewer = SetClipboardViewer(hwnd);
    return 0;
}

void
cleanup(HWND hwnd)
{
    remove_tasktray_icon(hwnd);
    ChangeClipboardChain(hwnd, hwndNextViewer);
}

LRESULT CALLBACK
WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    LRESULT ret = 0;
    static PAINTSTRUCT ps;
    HDC hdc;

    switch (uMsg) {
    case WM_CREATE:
	if (initialize(hwnd))
	    PostQuitMessage(0);
	break;

    case WM_PAINT:
	hdc = BeginPaint(hwnd, &ps);
	EndPaint(hwnd, &ps);
	break;

    case WM_CLOSE:
	cleanup(hwnd);
	PostQuitMessage(0);
	break;

    case WM_DESTROY:
	cleanup(hwnd);
	break;

    case WM_TASKTRAY:
	ret = TaskTrayProc(hwnd, wParam, lParam);
	break;

    case WM_COMMAND:
	switch (wParam) {
	case IDM_QUIT:
	    PostMessage(hwnd, WM_CLOSE, 0, 0);
	    break;
	}
	break;

    case WM_CHANGECBCHAIN:
	if ((HWND)wParam == hwndNextViewer)
	    hwndNextViewer = (HWND)lParam;
	else if (hwndNextViewer != NULL)
	    SendMessage(hwndNextViewer, uMsg, wParam, lParam);
	break;

    case WM_DRAWCLIPBOARD:
	DrawClipboardProc(hwnd, wParam, lParam);
	SendMessage(hwndNextViewer, uMsg, wParam, lParam);
	break;

    default:
	ret = DefWindowProc(hwnd, uMsg, wParam, lParam);
    }

    return ret;
}

static int
check_multiple(void)
{
    if (FindWindow(MODCLIP_WINDOW_CLASS, MODCLIP_WINDOW_NAME)) {
	MessageBox(NULL, "modclip is already executed.", "Error",
		   MB_OK|MB_ICONERROR);
	return -1;
    }
    return 0;
}

static int
register_class(void)
{
    WNDCLASS wc;

    memset(&wc, 0, sizeof (wc));
    wc.style = 0;
    wc.lpfnWndProc = &WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInst;
    wc.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_ICON));
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject(GRAY_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = MODCLIP_WINDOW_CLASS;
    if (!RegisterClass(&wc)) {
	MessageBox(NULL, "cannot register class", "Error", MB_OK|MB_ICONERROR);
	return -1;
    }
    return 0;
}

int
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
	int nCmdShow)
{
    HWND hwnd;
    MSG msg;

    hInst = hInstance;

    if (check_multiple())
	return 0;

    if (register_class())
	return 0;

    hwnd = CreateWindow(MODCLIP_WINDOW_CLASS, MODCLIP_WINDOW_NAME,
			WS_OVERLAPPEDWINDOW,
			CW_USEDEFAULT, CW_USEDEFAULT,
			100, 100,
			NULL,
			NULL,
			hInstance,
			NULL);

    /* main loop */
    while (GetMessage(&msg, NULL, 0, 0)) {
	TranslateMessage(&msg);
	DispatchMessage(&msg);
    }

    return 0;
}
